(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** Auxiliary language-related definitions *)

Require Import Arith Bool List Relations vgtac.
Require Export basis myclassical map countable.
Set Implicit Arguments.


Notation lookup := (list_assoc eqdec).

Lemma lookup2: forall K (M: EqDec K) X Y xsl xsr ysl ysr P (x: X) (y: Y) k
    (E: map fst ysl = map fst xsl)
    (INx: lookup k (xsl ++ xsr) = Some x)
    (INy: lookup k (ysl ++ ysr) = Some y)
    (Pl: forall k x y
             (INx: lookup k xsl = Some x)
             (INy: lookup k ysl = Some y),
           P x y : Prop)
    (Pr: forall k x y
             (INx: lookup k xsr = Some x)
             (INy: lookup k ysr = Some y),
           P x y),
    P x y.
Proof.
  i; apply list_assoc_app_or in INx; apply list_assoc_app_or in INy; des
  ; [|exfalso|exfalso|]; eauto using list_assoc_eqdom_false.
Qed.


(** Variables used in language expressions. *)
Definition termvar := nat.


(** Labels identifying module components. *)
Definition label := nat.


(** Events occuring in execution trace.

    [etau] is an internal (silent) event.
    [ein] and [eout] represent simple I/O.
*)
Inductive event :=
| etau
| ein (n: nat)
| eout (n: nat)
.


(** Binary operations on natural numbers. *)

Inductive binop := Plus | Minus | Mult | Equal.

Definition eqdec_binop (x y: binop) : sumbool (x = y) (x <> y).
  destruct x, y; vauto.
Defined.

Instance EqDec_binop : EqDec binop :=
  mkEqDec eqdec_binop.

Definition enum_binop (b: binop) : nat :=
  match b with
    | Plus => 0
    | Minus => 1
    | Mult => 2
    | Equal => 3
  end.

Instance Enumerable_binop: Enumerable binop :=
  {| enum := enum_binop |}.
  ii; destruct x, y; vauto.
Defined.

Definition decode_binop n : option binop :=
  match n with
    | 0 => Some Plus
    | 1 => Some Minus
    | 2 => Some Mult
    | 3 => Some Equal
    | _ => None
  end.

Lemma decode_binop_enum: forall r,
  decode_binop (enum r) = Some r.
Proof.
  destruct r; auto.
Qed.

Lemma decode_binop_some: forall n b
    (E: decode_binop n = Some b),
  n = enum_binop b.
Proof.
  i; repeat (destruct n; ss; vauto).
Qed.


Definition bosem (op: binop) (n m: nat) : nat :=
  match op with
  | Plus => n + m
  | Minus => n - m
  | Mult => n * m
  | Equal => if eqdec n m then 1 else 0
  end.


(** Name of a module's main function. *)
Definition MAIN := 42.
