(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_mid lang_tgt.
Require Import model_common model_untyped model_properties_untyped.
Require Import gw_common gw_untyped codegen codegen_correct_aux.
Require Import ssr util.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Lemma tm_refconfrel_write: forall (g : state (wlift mylw) -> vrelf) sg sl la lb va vb ha hb
    (RCR: tm_refconfrel (fun sg' => (g (sg', sl))) sg
            {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := {la |-> va} ## ha |}
            ({lb |-> vb} ## hb, Nothing))
    (RL: snd sg la lb)
    va' vb' (VCL: vclos tm_gw sg (g (sg, sl)) va' vb'),
  tm_refconfrel (fun sg' => (g (sg', sl))) sg
    {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := {la |-> va'} ## ha |}
    ({lb |-> vb'} ## hb, Nothing).
Proof.
  i; rrdes RCR; clarify; rr; esplits; eauto using pmvalid_plus_cell.
  - i; dup ALC; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; subst; eauto.
    + hexploit DOMa; [|by eauto].
      rrdes ALC0; eapply pmvalid_plus_cell in VLD.
      eapply pmread_plus; eauto; pmplus_eq'.
  - i; dup ALC; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; subst; eauto.
    + hexploit DOMb; [|by eauto].
      rrdes ALC0; eapply pmvalid_plus_cell in VLD.
      eapply pmread_plus; eauto; pmplus_eq'.
  - i; dup REG; apply RELV in REG; des.
    dup RDa; rrdes RDa; dup RDb; rrdes RDb.
    apply pmread_plusD in RDa0; apply pmread_plusD in RDb0; des.
    + apply pmread_cellD in RDa0; apply pmread_cellD in RDb0; des; subst.
      esplits; last (exact VCL)
      ; eauto using pmread_plus, pmread_cell, pmvalid_plus_cell.
    + exfalso.
      apply pmread_cellD in RDb0; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X0 eq_refl). subst.
      rrdes RDa0; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLD.
      apply pmvalid_plus_l in VLD.
      apply pmvalid_cell_cell in VLD; vauto.
    + exfalso.
      apply pmread_cellD in RDa0; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X eq_refl). subst.
      rrdes RDb0; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLD0.
      apply pmvalid_plus_l in VLD0.
      apply pmvalid_cell_cell in VLD0; vauto.
    + eauto 10 using pmread_plus, pmplus_comm, pmvalid_plus_cell.
Qed.

Lemma tr_exp_correct_write: forall G x y e
    (INx: In (inl x) G)
    (INy: In (inl y) G)
    (IHe: tm_psim G
            (fun imports => map enc (tr_exp e imports (tvars G) (kvars G)))
            e),
  tm_psim G (fun imports => map enc (tr_exp (IEwrite x y e) imports (tvars G) (kvars G)))
            (IEwrite x y e).
Proof.
  i; hexploit getVal_length; i; des.
  intro; specialize (IHe i); rrdes IHe.
  eexists; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV INx); hexploit (values_for_var RSV RDV INy).
  i; des.

  tm_prepare.

  destruct (LEM (exists locb, vb = IVloc locb))
  ; last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    eapply esim_err.
    - instantiate (1:=(None, (envb, IEwrite x y e))). 
      econs 1; eapply IRerr; vauto; i; invc FH.
      rewrite INb in GET; clarify; eauto 10.
    - eauto.
  }
  des; subst.

  
  apply tm_vclosDb in VCL; des; rrdes VCL; clarify; clean.
  dup WHY0; rrdes WHY0; hexploit RELV; eauto; i; des; clarify.
  rrdes RDa RDb; subst; rewrite <-!pmplus_assoc in *.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits; rewrite <-!pmplus_assoc in *.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRwrite; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
      + rewrite pmplus_assoc; eauto.
      + rewrite -pmplus_assoc; eauto using pmvalid_plus_cell.
      + rewrite -pmplus_assoc; eauto using pmfinite_plus_cell.
  }

  do_getVal x Eladdr X.
  do_getVal y Eladdr X.

  eapply pesim_dance.
  - rrdes RLa0; instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto.
  - clear SV HV FIN; rename mrf into mrf0.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr !map_app !pmcells_app !map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq'.
      - rewrite (pmplus_swap hp).
        
        
        
        hexploit (sto_hind (mpc + length (getVal x senva (tvars G) r_clo) +
                            length (getVal y senva (tvars G) r_arg)) mrf
                           (freest ## fst s4 ## cst) (hp ## h' ## snd s4 ## chp)
                           r_clo o_zero va1 (mrf r_arg)).
        unfold mrf; s; tmext_simpl; eauto.
    }
    { rewrite (pmplus_swap hp) in HV0, FIN0.
      depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell.
    }

    eapply dance_O; instantiate (1 := fun _ => _); s; eauto.

    - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
      esplits; first 2 last; last first.
      + right; splits; eauto using wfo_le_n_add.
      + i; frames_nodep RLa cfas.
        destruct ea'; clear SV HV FIN; destruct_tgt_real RLa.

        setcrf mrf'.
        set (s' := (mrf', P0, s, s3, s4)).
        assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
        { clear -HV; unfold s'.
          s; splits; tmext_simpl; auto.
        }
        exists s'; esplits; [exact EXT'|].

        rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
        { unfold s'; rr; esplits.
          - rr; esplits; eauto.
            + s; rr; esplits; eauto; clean.
              instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp |}).
              rrdes CPa; s in HP.
              unfold mrf'; rr; splits; s; tmext_simpl; auto.
            + s; inst_pairs; eapply tm_refconfrel_write; eauto.
              * eapply tm_refconfrel_mon
                ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
                ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
              * eapply vclos_unlift, vclos_mon; eauto using (gk_mon GK).
          - rr; esplits; eauto.
          - s; noemp; f_equal.
            lhs rewrite pmplus_swap -!pmplus_assoc; eauto.
          - noemp; auto.
        }

        inst_pairs; eapply IHe; eauto using ext_trans.
        * i; hexploit RSV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * i; hexploit RDV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * unfold s', mrf'; s; tmext_simpl; eauto.
        * unfold s', mrf'; s; tmext_simpl; eauto.
        * rewrite -(app_nil_l envb).
          eapply tm_ksim_mon; eauto using ext_trans.
          unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
        * s; econs; symmetry.
          rewrite Eladdr !map_app !pmcells_app !map_length !pmcells_cons.
          pmplus_eq'.
      + rr; esplits; s; noemp; eauto.
Grab Existential Variables.
exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0).
Qed.



Lemma tr_exp_correct_new: forall G x bx e
    (IN: In (inl x) G)
    (IHe: tm_psim (inl (inr bx) :: G)
            (fun imports => map enc (tr_exp e imports (bx :: tvars G) (kvars G)))
            e),
  tm_psim G
    (fun imports => map enc (tr_exp (IEnew x bx e) imports (tvars G) (kvars G)))
    (IEnew x bx e).
Proof.
  i; hexploit getVal_length; i; des.
  intro; specialize (IHe i); rrdes IHe.
  econs; i; destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV IN); i; des.
  s in Eladdr.

  tm_prepare.

  

  do_getVal x Eladdr X.

  
  pfold; econs; i; esplits; eauto.
  destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
  destruct cfb; rrdes XTRb; clarify.
  invc CNF; des; subst; noemp.
  rrdes MB; s in Ec; clarify.
  destruct ma; clear SV HV FIN; simpl cplus in MA.
  destruct_tgt_real MA.
  rewrite <-!pmplus_assoc in*. clarify.

  eapply esim_step; first (i; rrdes FH; ss; omega).
  i; s in ST; eapply tgt_det in ST; first 1 last; [|by eauto|by eauto| |].
  { eapply Sok; [ii; ss; omega|eapply Sld].
    - eapply pmread_plus; [by eapply pmread_cell| |by eauto].
      rewrite Eladdr cons_app3 !map_app !pmcells_app !map_length.
      rewrite !pmcells_cons; s; do 2 pmplus_eq''.
    - eauto.
  }
  { ii; vauto. }
  des; subst.

  dup PV; eapply (pmfinite_alloc vb) in PV; eauto; des.
  esplits; first (by eauto); [left; econs; eauto| |by eauto]
  ; last first; last 1 first.
  { left; inst_pairs; eapply IRok; eapply IRnew; eauto. }
  { econs; eauto.
    + instantiate (1 := mkconf _ _ _ _); s; noemp.
      rr; esplits; s; eauto; pmplus_eq.
    + inst_pairs; s; noemp; rewrite pmplus_assoc.
      rr; esplits; eauto; rewrite -pmplus_assoc; eauto.
  }
  rewrite clear_bote; eapply gres_incl_id.

  eapply pesim_dance.
  - rrdes RLa0; instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto.
  - clear SV HV FIN FIN0; rename mrf into mrf0.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - rewrite !pmcells_cons -!pmplus_assoc.
        replace newloc with (add_offset (mrf r_clo) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }
    tmext_uniq.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; fold newloc; rewrite ->(pmplus_comm chp3), <-!pmplus_assoc. 
    setnewloc.
    tmext_uniq.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - rewrite !pmcells_cons -!pmplus_assoc (pmplus_swap {_|->_}).
        replace (S newloc0) with (add_offset (mrf r_arg) o_one)
        ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - rewrite !pmcells_cons -!pmplus_assoc.
        replace newloc0 with (add_offset (mrf r_arg) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }             

    eapply dance_O; s.
    instantiate (1 := fun cf _ _ cf' _ _ _ => depmach2 cf cf'); s; eauto.

  - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 0 RLa0 XTRa cfas; frames_dep 2 RLa XTRa cfas
      ; frames_nodep RLa0 cfas.
      destruct ea'; clear SV HV FIN; destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setnewloc.

      setcrf mrf'.
      assert (VLDa: pmvalid ({newloc0 |-> va} ## refha)).
      { rewrite pmcells_sng; decomp_pmvalid HV; eauto. }
      assert (VLDb: pmvalid ({x0 |-> vb} ## refhb)).
      { decomp_pmvalid PV; eauto. }
      set (s' := (mrf', P0, s,
                  refst_allocP s3 newloc0 x0 (tm_refconfrel_newa WHY0 VLDa) (tm_refconfrel_newb WHY0 VLDb),
                  (fst s4, {newloc |=> [newloc0; mrf r_env]} ## snd s4))).
      assert (VLD': pmvalid ({newloc |=> [newloc0; mrf r_env]} ## snd s4)).
      { decomp_pmvalid HV; eauto. }
      assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'; unfold s'.
        unfold refst_alloc; s; splits; tmext_simpl; auto.
        - unfold refst_alloc; eauto.
        - i; eapply pmread_plus; eauto; pmplus_eq'.
      }
      exists s'; esplits; [exact EXT'|].

      rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s'; rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto; clean.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp |}).
            rrdes CPa; s in HP; unfold mrf'; rr; splits; s; tmext_simpl; auto.
          + inst_pairs; eapply tm_refconfrel_alloc; s; eauto.
        - rr; esplits; eauto.
        - s; noemp; f_equal; lhs rewrite pmplus_swap (pmplus_swap hp).
          rewrite pmcells_sng -2!pmplus_assoc; eauto.
        - noemp; auto.
      }

      inst_pairs; eapply IHe; eauto using ext_trans.
      * instantiate (1 := (inl (inr bx), IVloc x0) :: envb).
        instantiate (1 := senva).
        i; apply in_consD in IN0; des; vauto.
        rewrite list_assoc_cons_neq; [|by ii; vauto].
        hexploit RSV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * instantiate (1 := (bx, newloc0) :: denva).
        i; apply in_consD in IN0; des.
        { clarify; esplits; eauto.
          econs; eapply vclgen_ref; econs; unfold s', refst_allocP, refst_alloc
          ; s; eauto 10.
        }
        rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN1; vauto]. 
        rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN1; vauto]. 
        hexploit RDV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * unfold s', mrf'; s; tmext_simpl; esplits
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
        eapply reprlist_mon; eauto.
        i; eapply pmread_plus; eauto; pmplus_eq'.
      * s; rewrite ORD; auto.
      * unfold s', mrf'; s; tmext_simpl; eauto.
      * s; rewrite cons_app.
        eapply tm_ksim_mon; eauto using ext_trans.
        unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
      * s; econs; symmetry.
        rewrite Eladdr !pmcells_cons !map_app !pmcells_app !map_length; s.
        rewrite !pmcells_cons; s; pmplus_eq'.
      + rr; esplits; s; noemp; eauto.
Grab Existential Variables.
exact (wfo_nat 0).
Qed.



Lemma tr_exp_correct_input: forall G bx e
    (IHe: tm_psim (inl (inr bx) :: G)
            (fun imports => map enc (tr_exp e imports (bx :: tvars G) (kvars G)))
            e),
  tm_psim G (fun imports => map enc (tr_exp (IEinput bx e) imports (tvars G) (kvars G)))
            (IEinput bx e).
Proof.
  i; eexists; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  specialize (IHe i); rrdes IHe.

  tm_prepare.

  
  pfold; econs; i; esplits; eauto.
  destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
  destruct cfb; rrdes XTRb; clarify.
  invc CNF; des; subst; noemp.
  rrdes MB; s in Ec; clarify.
  destruct ma; clear SV HV FIN; simpl cplus in MA.
  destruct_tgt_real MA.
  rewrite <-!pmplus_assoc in*. clarify.


  assert (INP: pmread (hp ## refha ## snd s4 ## chp) mpc0 (enc (i_inp r_aux))).
  { read_code Eladdr. }

  eapply esim_step; first (by eauto); i.
  inv ST; last by (exfalso; eapply ST0, Sinp; eauto).
  inv ST0; hexploit (pmread_uniq INS INP); intro ENC
  ; apply enc_inj in ENC; vauto.
  esplits; first (by eauto); last (by auto); [left; econs; eauto|]
  ; last first; last 1 first.
  { left; inst_pairs; eapply IRok; eauto. }
  { econs; eauto.
    + instantiate (1:=mkconf _ _ _ _); s; noemp; eauto.
      rr; esplits; s; eauto; pmplus_eq.
    + inst_pairs; s; noemp; eauto.
      rr; esplits; s; eauto; pmplus_eq.
  }
  rewrite clear_bote; eapply gres_incl_id.
  
  eapply pesim_dance.
  - rrdes RLa0; instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto.
  - clear SV HV FIN FIN0 ST ST0 NZ NZ0 INS INP MA0 RLa0.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld]; [by read_code Eladdr|by eauto]. }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc]; [by read_code Eladdr|by eauto]. }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto]; [by read_code Eladdr|].
      rewrite !pmcells_cons -!pmplus_assoc.
      replace newloc with (add_offset (mrf r_arg) o_zero)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }
    s; fold newloc.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto]; [by read_code Eladdr|].
      rewrite !pmcells_cons -!pmplus_assoc (pmplus_swap {_|->_}).
      replace (S newloc) with (add_offset (mrf r_arg) o_one)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld]; [by read_code Eladdr|by eauto]. }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := fun _ cf _ _ _ => depmach1 cf); s; eauto.

  - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.  
    + i; frames_dep 1 RLa XTRa cfas.
      frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN ST.
      destruct_tgt_real RLa.
      setnewloc.

      setcrf mrf'.
      set (s' := (mrf', P0, s, s3, (fst s4, {newloc |=> [n; mrf0 r_env]} ## snd s4))).
      assert (VLD': pmvalid ({newloc |=> [n; mrf0 r_env]} ## snd s4)).
      { decomp_pmvalid HV; eauto. }
      assert (EXT': ext (wlift mylw) s' (mrf0, P0, s, s3, s4)).
      { clear -HV VLD'; unfold s'.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto; pmplus_eq'.
      }
      exists s'; esplits; [exact EXT'|].

      rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s'; rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto; clean.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp |}).
            rrdes CPa; s in HP.
            unfold mrf'; rr; splits; s; tmext_simpl; auto.
          + eapply tm_refconfrel_mon
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
        - rr; esplits; eauto.
        - s; noemp; f_equal.
          lhs rewrite pmplus_swap -pmplus_assoc; eauto.
        - noemp; auto.
      }                         

      inst_pairs; eapply IHe; eauto using ext_trans.
      * instantiate (1 := (inl (inr bx), IVnat n) :: envb).
        instantiate (1 := senva).
        i; apply in_consD in IN; des; vauto.
        rewrite list_assoc_cons_neq; [|by ii; vauto].
        hexploit RSV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * instantiate (1 := (bx, n) :: denva).
        i; apply in_consD in IN; des.
        { clarify; esplits; eauto.
          econs; eapply vclgen_nat; s; eauto.
          rewrite plus_0_r; eauto.
        }
        rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN0; vauto]. 
        rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN0; vauto]. 
        hexploit RDV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK). 
      * unfold s', mrf'; s; tmext_simpl; esplits
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
        eapply reprlist_mon; eauto.
        i; eapply pmread_plus; eauto; pmplus_eq'.
      * s; rewrite ORD; auto.
      * unfold s', mrf'; s; tmext_simpl; eauto.
      * s; rewrite cons_app.
        eapply tm_ksim_mon; eauto using ext_trans.
        unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
      * s; econs; symmetry.
        rewrite Eladdr !pmcells_cons; s; pmplus_eq'.
    + rr; esplits; s; noemp; eauto.
Grab Existential Variables.
exact 0. exact (wfo_nat 0).
Qed.



Lemma tr_exp_correct_output: forall G x e
    (IN: In (inl x) G)
    (IHe: tm_psim G
            (fun imports => map enc (tr_exp e imports (tvars G) (kvars G)))
            e),
  tm_psim G (fun imports => map enc (tr_exp (IEoutput x e) imports (tvars G) (kvars G)))
            (IEoutput x e).
Proof.
  i; hexploit getVal_length; i; des.
  eexists; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV IN); i; des.
  specialize (IHe i); rrdes IHe.

  tm_prepare.

  destruct (LEM (exists n, vb = IVnat n)); last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    eapply esim_err.
    - instantiate (1:=(None, (envb, IEoutput x e))).
      econs 1; eapply IRerr; vauto; i; invc FH.
      rewrite INb in GET; clarify; eauto 10.
    - eauto.
  }                             
  des; subst.

  do_getVal x Eladdr X.

  dup VCL; apply tm_vclosDb in VCL; des; s in VCL; subst.

  
  pfold; econs; i; esplits; eauto.
  destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
  destruct cfb; rrdes XTRb; clarify.
  invc CNF; des; subst; noemp.
  rrdes MB; s in Ec; clarify.
  destruct ma; clear SV HV FIN; simpl cplus in MA.
  destruct_tgt_real MA.
  rewrite <-!pmplus_assoc in*. clarify.

  assert (OUT: pmread (hp ## refha ## snd s4 ## chp)
                      (mpc + length (getVal x senva (tvars G) r_clo))
                      (enc (i_out r_clo))).
  { eapply pmread_plus; eauto.
    rewrite Eladdr map_app pmcells_app; s.
    rewrite !pmcells_cons map_length; pmplus_eq'.
  }

  eapply esim_step; first (i; rrdes FH; ss; omega); i.
  inv ST; last by (exfalso; eapply ST0, Sout; eauto).
  inv ST0; hexploit (pmread_uniq INS OUT); intro ENC
  ; apply enc_inj in ENC; vauto.

  set (s' := ([n0//r_clo]>>mrf, P0, s, s3, s4)).
  assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
  { rr; s; esplits; eauto. }

  esplits; first (by eauto); last (exact EXT'); [left; econs; eauto|]
  ; last first; last 1 first.
  { left; tmext_simpl; inst_pairs; eapply IRok; eauto. }
  { econs; eauto.
    + instantiate (1:=mkconf _ _ _ _); s; noemp; eauto.
      rr; esplits; s; eauto; pmplus_eq.
    + inst_pairs; s; noemp; eauto.
      rr; esplits; s; eauto; pmplus_eq.
  }
  rewrite clear_bote; eapply gres_incl_id.

  rewrite tgt_csplit mid_csplit; eapply pesim_release; last first; try by rr; s; eauto.
  { unfold s'; rr; esplits.
    - rr; esplits; eauto.
      + s; rr; esplits; eauto; clean.
        instantiate (1 := {| cpc := Nothing; crf := Just [n0//r_clo]>>mrf; cst := freest; chp := hp |}).
        rrdes CPa; s in HP.
        rr; splits; s; tmext_simpl; auto.
      + eapply tm_refconfrel_mon
        ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
        ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
    - rr; esplits; eauto.
    - s; noemp; f_equal.
      lhs rewrite -pmplus_assoc; eauto.
    - noemp; auto.
  }

  inst_pairs; eapply IHe; eauto using ext_trans.
  - i; hexploit RSV; eauto; i; des; esplits; eauto.
    eapply vclos_mon; eauto using (gk_mon GK).
  - i; hexploit RDV; eauto; i; des; esplits; eauto.
    eapply vclos_mon; eauto using (gk_mon GK).
  - unfold s'; s; tmext_simpl; eauto.
  - unfold s'; s; tmext_simpl; eauto.
  - rewrite -(app_nil_l envb).
    eapply tm_ksim_mon; eauto using ext_trans.
    unfold s'; s; tmext_simpl; rewrite SP; omega.
  - s; econs; symmetry.
    rewrite Eladdr !map_app !pmcells_app !map_length; s.
    rewrite !pmcells_cons; pmplus_eq'.
Grab Existential Variables.
exact (wfo_nat 0).
Qed.



Lemma tr_exp_correct_read: forall G x bx e
    (IN: In (inl x) G)
    (IHe: tm_psim (inl (inr bx) :: G)
            (fun imports => map enc (tr_exp e imports (bx :: tvars G) (kvars G)))
            e),
  tm_psim G
    (fun imports => map enc (tr_exp (IEread x bx e) imports (tvars G) (kvars G)))
    (IEread x bx e).
Proof.
  i; hexploit getVal_length; i; des.
  intro; specialize (IHe i); rrdes IHe.
  econs; i; destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV IN); i; des.
  s in Eladdr.

  tm_prepare.

  destruct (LEM (exists loc, vb = IVloc loc))
  ; last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    eapply esim_err.
    - instantiate (1:=(None, (envb, IEread x bx e))).
      econs 1; eapply IRerr; vauto; i; invc FH
      ; rewrite INb in GET; clarify; eauto 10.
    - eauto.
  }
  des; subst.

  do_getVal x Eladdr X.

  
  apply tm_vclosDb in VCL; des; rrdes VCL; clarify; clean.
  dup WHY0; rrdes WHY0; hexploit RELV; eauto; i; des.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRread; s; eauto using pmread_plus.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0.
  - clear SV HV FIN; rename mrf into mrf0.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - rewrite !pmcells_cons -!pmplus_assoc (pmplus_swap {_|->_}).
        replace (S newloc) with (add_offset (mrf r_clo) o_one)
        ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - econs; eapply pmread_plus; unfold mrf; s; tmext_simpl
        ; first (eapply RDa); eauto; pmplus_eq'.
    }
    { depmcplus0. }
    tmext_uniq.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - rewrite !pmcells_cons -!pmplus_assoc.
        replace (newloc) with (add_offset (mrf r_clo) o_zero)
        ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }

    eapply dance_O; s.
    instantiate (1 := fun _ cf _ _ _ _ => depmach1 cf); s; eauto.

  - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN; destruct_tgt_real RLa.
      setnewloc.

      setcrf mrf'.
      set (s' := (mrf', P0, s, s3, (fst s4, {newloc |=> [va0; mrf r_env]} ## snd s4))).
      assert (VLD': pmvalid ({newloc |=> [va0; mrf r_env]} ## snd s4)).
      { decomp_pmvalid HV; eauto. }
      assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'; unfold s'.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto; pmplus_eq'.
      }
      exists s'; esplits; [exact EXT'|].

      rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s'; rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto; clean.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp |}).
            rrdes CPa; s in HP; unfold mrf'; rr; splits; s; tmext_simpl; auto.
          + eapply tm_refconfrel_mon
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
        - rr; esplits; eauto.
        - s; noemp; f_equal; lhs rewrite pmplus_swap -pmplus_assoc; eauto.
        - noemp; auto.
      }

      inst_pairs; eapply IHe; eauto using ext_trans.
      * instantiate (1 := (inl (inr bx), vb) :: envb).
        instantiate (1 := senva).
        i; apply in_consD in IN0; des; vauto.
        rewrite list_assoc_cons_neq; [|by ii; vauto].
        hexploit RSV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * instantiate (1 := (bx, va0) :: denva).
        i; apply in_consD in IN0; des.
        { clarify; esplits; eauto.
          eapply vclos_mon; first (eapply vclos_lift); eauto using (gk_mon GK).
        }
        rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN1; vauto]. 
        rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN1; vauto]. 
        hexploit RDV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * unfold s', mrf'; s; tmext_simpl; esplits
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
        eapply reprlist_mon; eauto.
        i; eapply pmread_plus; eauto; pmplus_eq'.
      * s; rewrite ORD; auto.
      * unfold s', mrf'; s; tmext_simpl; eauto.
      * s; rewrite cons_app.
        eapply tm_ksim_mon; eauto using ext_trans.
        unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
      * s; econs; symmetry.
        rewrite Eladdr !pmcells_cons !map_app !pmcells_app !map_length; s.
        rewrite !pmcells_cons; s; pmplus_eq'.
      + rr; esplits; s; noemp; eauto.
Grab Existential Variables.
exact (wfo_nat 0).
Qed.



Lemma tr_exp_correct_app: forall G f x k
    (INf: In (inl f) G)
    (INx: In (inl x) G)
    (INk: In (inr k) G),
  tm_psim G (fun imports => map enc (tr_exp (IEapp f x k) imports (tvars G) (kvars G)))
            (IEapp f x k).
Proof.
  i; hexploit getVal_length; i; des.
  eexists; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV INx); hexploit (values_for_var RSV RDV INf).
  i; des; clear RSV RDV.
  eapply kvars_in in INk; hexploit tm_ksimD; eauto; i; des.

  tm_prepare.

  destruct (LEM (exists env bf bx bk e, vb0 = IVclo env (IEfix bf bx bk e)))
  ; last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    eapply esim_err.
    - instantiate (1:=(None, (envb, IEapp f x k))).
      econs 1; eapply IRerr; vauto; i; invc FH.
      rewrite INb0 in GET; clarify; eauto 10.
    - eauto.
  }
  des; subst.

  do_getVal x Eladdr X.
  do_getVal f Eladdr X.

  eapply pesim_getCont; first omega.
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr cons_app3 !map_app !pmcells_app !map_length; s; pmplus_eq'.
  }
  { s; eapply wfo_le_n_add. }
  { tmext_simpl; eapply pmread_plus; eauto; pmplus_eq'. }
  { rewrite plus_comm; tmext_simpl; eapply pmread_plus; eauto; pmplus_eq'. }
  { tmext_simpl; rewrite SP.
    instantiate (1 := 2 * length (kvars G) + r r_sp - 2 * (1 + index k (kvars G))).
    hexploit (list_index_some _ _ INk); i; des.
    rewrite X0; s; eapply list_index_length in X0; omega.
  }

  
  dup VCL0; apply tm_vclosDb in VCL0; des; dup VQ; rrdes VQ.
  dup CPa; rrdes CPa; clean; dup VQ; apply HP in VQ; s in VQ; clear HP ST.

  
  pfold; econs; i; esplits; eauto.
  destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
  destruct cfb; rrdes XTRb; clarify.
  invc CNF; des; subst; noemp.
  rrdes MB; s in Ec; clarify.
  destruct ma; clear SV HV FIN; simpl cplus in MA.
  destruct_tgt_real MA.
  rewrite <-!pmplus_assoc in*. clarify.

  eapply esim_step; first (i; rrdes FH; ss; omega).
  i; s in ST; eapply tgt_det in ST; first 1 last; [|by eauto|by eauto| |].
  { eapply Sok; [ii; ss; omega|eapply Sjmp].
    + eapply pmread_plus; [by eapply pmread_cell| |by eauto].
      rewrite Eladdr cons_app3 !map_app !pmcells_app !map_length.
      rewrite !pmcells_cons; s; pmplus_eq''.
    + econs; tmext_simpl; s; eapply pmread_plus; eauto.
  }
  { ii; vauto. }
  des; subst.

  hexploit (list_index_some _ _ INk); i; des.
  rewrite ->X0 in *; simpl withd in *; rename n0 into idx.
  eapply list_index_length in X0.
  hexploit stack_split; eauto; i; des.
  hide $ SPL POPPED USED.
  rewrite SPL pmcells_app -pmplus_assoc in LST.

  setmrf mrf'.
  set (s' := (mrf', P0, s, s3, ({r r_sp |=> used} ## oldst, snd s4))).
  assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
  { unfold s', mrf'; s; splits; tmext_simpl; auto. }

  do 6 econs; exists s'.
  esplits; first (by eauto); [right; econs; eauto| |]; last first; last 1 first. 
  - inst_pairs; eapply IRok; eauto using IRapp.
  - s; eauto.
  - econs; first 1 last; last first.
    { rr; esplits; last by eauto.
      - rr; esplits; [ | |by eauto|by eauto]; last first.
        + eapply tm_refconfrel_mon
          ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
          ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
        + rr; splits; last by eauto.
          eapply tgt_confpred_pop with (popped:=popped); eauto.
          { unfold mrf'; tmext_simpl; rewrite SP LF; s; eauto. }
          { replace (mrf' r_sp) with (r r_sp + length used)
            ; last (unfold mrf'; tmext_simpl; omega).
            rewrite pmplus_comm; rewrite LST in SV; decomp_pmvalid SV; eauto. }
      - rr; esplits; eauto.
      - s; noemp; eauto.
    }
    { instantiate (1:=(pmemp, _)). noemp. s. rr; esplits; s; noemp; eauto. }
    { instantiate (1 := mkconf (Just _) Nothing pmemp pmemp). s.
      rewrite <-!pmplus_assoc. noemp.
      unfold mrf'; tmext_simpl.
      rr; esplits; s; noemp; eauto.
      rewrite -(pmplus_swap freest) LST -!pmplus_assoc (pmplus_swap {_|=>_}).
      rewrite -(plus_comm (r r_sp)) LU ; s; do 4 f_equal; omega.
    }
    { split; rr; s; eauto. }
  - eapply call_app; last first; first 1 last.
    + move VCL at bottom; eapply vclos_mon; eauto using (gk_mon GK).
    + eapply vclf_fix; s; eauto using (gk_mon GK).
    + s; eauto.
    + s; replace va0 with (mrf' r_clo); last (by unfold mrf'; tmext_simpl; auto).
      replace va with (mrf' r_arg); last (by unfold mrf'; tmext_simpl; auto).
      econs; s; unfold mrf'; tmext_simpl; eauto.
    + unfold mrf'; clean; tmext_simpl.
      econs; i; rrdes EXTP EXTP EXTP EXTP; rrdes EXTP0.
      
      rewrite clear_bote; eapply gres_incl_id, pesim_gknow; i; eapply ESIM; eauto.
      * rewrite -RF SP; unfold s', mrf'; s; tmext_simpl; auto.
      * rewrite ST LST; unfold s'; s; rewrite -!pmplus_assoc -RF.
        unfold s', mrf'; s; tmext_simpl.
        rewrite (pmplus_comm _ oldst) LU; do 3 f_equal; omega.
      * rewrite -RF0; unfold s', mrf'; s; tmext_simpl; auto.
Grab Existential Variables.
  exact wfo_zero. exact wfo_zero.
Qed.



Lemma tr_exp_correct_inst: forall G f k
    (INf: In (inl f) G)
    (INk: In (inr k) G),
  tm_psim G (fun imports => map enc (tr_exp (IEinst f k) imports (tvars G) (kvars G)))
            (IEinst f k).
Proof.
  i; hexploit getVal_length; i; des.
  eexists; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV INf); i; des; clear RSV RDV.
  eapply kvars_in in INk; hexploit tm_ksimD; eauto; i; des.

  tm_prepare.

  destruct (LEM (exists env bk e, vb = IVclo env (IEgen bk e)))
  ; last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    eapply esim_err.
    - instantiate (1:=(None, (envb, IEinst f k))).
      econs 1; eapply IRerr; vauto; i; invc FH.
      rewrite INb in GET; clarify; eauto 10.
    - eauto.
  }
  des; subst.

  do_getVal f Eladdr X.

  eapply pesim_getCont; first omega.
  { i; eapply pmread_plus; eauto.
    rewrite Eladdr cons_app3 !map_app !pmcells_app !map_length; s; pmplus_eq'.
  }
  { s; eapply wfo_le_n_add. }
  { tmext_simpl; eapply pmread_plus; eauto; pmplus_eq'. }
  { rewrite plus_comm; tmext_simpl; eapply pmread_plus; eauto; pmplus_eq'. }
  { tmext_simpl; rewrite SP.
    instantiate (1 := 2 * length (kvars G) + r r_sp - 2 * (1 + index k (kvars G))).
    hexploit (list_index_some _ _ INk); i; des.
    rewrite X0; s; eapply list_index_length in X0; omega.
  }

  
  dup VCL; apply tm_vclosDb in VCL; des; dup VQ; rrdes VQ.
  dup CPa; rrdes CPa; clean; dup VQ; apply HP in VQ; s in VQ; clear HP ST.

  
  pfold; econs; i; esplits; eauto.
  destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
  destruct cfb; rrdes XTRb; clarify.
  invc CNF; des; subst; noemp.
  rrdes MB; s in Ec; clarify.
  destruct ma; clear SV HV FIN; simpl cplus in MA.
  destruct_tgt_real MA.
  rewrite <-!pmplus_assoc in*. clarify.

  eapply esim_step; first (i; rrdes FH; ss; omega).
  i; s in ST; eapply tgt_det in ST; first 1 last; [|by eauto|by eauto| |].
  { eapply Sok; [ii; ss; omega|eapply Sjmp].
    + eapply pmread_plus; [by eapply pmread_cell| |by eauto].
      rewrite Eladdr cons_app3 !map_app !pmcells_app !map_length.
      rewrite !pmcells_cons; s; pmplus_eq''.
    + econs; tmext_simpl; s; eapply pmread_plus; eauto.
  }
  { ii; vauto. }
  des; subst.

  hexploit (list_index_some _ _ INk); i; des.
  rewrite ->X0 in *; simpl withd in *; rename n0 into idx.
  eapply list_index_length in X0.
  hexploit stack_split; eauto; i; des.
  hide $ SPL POPPED USED.
  rewrite SPL pmcells_app -pmplus_assoc in LST.

  setmrf mrf'.
  set (s' := (mrf', P0, s, s3, ({r r_sp |=> used} ## oldst, snd s4))).
  assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
  { unfold s', mrf'; s; splits; tmext_simpl; auto. }

  do 6 econs; exists s'.
  esplits; first (by eauto); [right; econs; eauto| |]; last first; last 1 first. 
  - inst_pairs; eapply IRok; eauto using IRinst.
  - s; eauto.
  - econs; first 1 last; first last.
    { rr; esplits; last by eauto.
      - rr; esplits; [ | |by eauto|by eauto]; last first.
        + eapply tm_refconfrel_mon
          ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
          ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
        + rr; splits; last by eauto.
          eapply tgt_confpred_pop with (popped:=popped); eauto.
          { unfold mrf'; tmext_simpl; rewrite SP LF; s; eauto. }
          { replace (mrf' r_sp) with (r r_sp + length used)
            ; last (unfold mrf'; tmext_simpl; omega).
            rewrite pmplus_comm; rewrite LST in SV; decomp_pmvalid SV; eauto. }
      - rr; esplits; eauto.
      - s; noemp; eauto.
    }
    { instantiate (1:=(pmemp, _)). noemp. s. rr; esplits; s; noemp; eauto. }
    { instantiate (1 := mkconf (Just _) Nothing pmemp pmemp). s.
      rewrite <-!pmplus_assoc. noemp.
      unfold mrf'; tmext_simpl.
      rr; esplits; s; noemp; eauto.
      rewrite -(pmplus_swap freest) LST -!pmplus_assoc (pmplus_swap {_|=>_}).
      rewrite -(plus_comm (r r_sp)) LU ; s; do 4 f_equal; omega.
    }
    { split; rr; s; eauto. }
  - eapply call_inst; last first; first 1 last.
    + eapply vclf_gen; s; eauto using (gk_mon GK).
    + s; eauto.
    + s; replace va with (mrf' r_clo); last (by unfold mrf'; tmext_simpl; auto).
      econs; s; unfold mrf'; tmext_simpl; eauto.
    + unfold mrf'; clean; tmext_simpl.
      econs; i; rrdes EXTP EXTP EXTP EXTP; rrdes EXTP0.
      rewrite clear_bote; eapply gres_incl_id, pesim_gknow; i; eapply ESIM; eauto.
      * rewrite -RF SP; unfold s', mrf'; s; tmext_simpl; auto.
      * rewrite ST LST; unfold s'; s; rewrite -!pmplus_assoc -RF.
        unfold s', mrf'; s; tmext_simpl.
        rewrite (pmplus_comm _ oldst) LU; do 3 f_equal; omega.
      * rewrite -RF0; unfold s', mrf'; s; tmext_simpl; auto.
Grab Existential Variables.
  exact wfo_zero.
Qed.



Lemma tr_exp_correct_ifnz: forall G x el er
    (IN: In (inl x) G)
    (IHel: tm_psim G
            (fun imports => map enc (tr_exp el imports (tvars G) (kvars G)))
            el)
    (IHer: tm_psim G
            (fun imports => map enc (tr_exp er imports (tvars G) (kvars G)))
            er),
  tm_psim G
    (fun imports => map enc (tr_exp (IEifnz x el er) imports (tvars G) (kvars G)))
    (IEifnz x el er).
Proof.
  i; hexploit getVal_length; i; des.
  econs; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  specialize (IHel i); rrdes IHel.
  specialize (IHer i); rrdes IHer.
  hexploit (values_for_var RSV RDV IN); i; des.

  tm_prepare.

  destruct (LEM (vb = IVnat 0 \/ exists n, vb = IVnat (S n)))
  ; last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    eapply esim_err.
    - instantiate (1:=(None, (envb, IEifnz x el er))).
      econs 1; eapply IRerr; vauto; i; invc FH
      ; rewrite INb in GET; clarify; eauto 10.
    - eauto.
  }
  des; subst.

  { 

    eapply pesim_stepb_obsolete.
    { i; rrdes XTR RLb; clarify.
      rrdes RL; s in Ec; noemp; clarify; esplits.
      - econs 1; inst_pairs; eapply IRok.
        eapply IRifz; s; rewrite INb; s; eauto.
      - inst_pairs; s; noemp; rr; esplits; eauto.
    }

    do_getVal x Eladdr X.

    dup VCL; apply tm_vclosDb in VCL; des; rrdes VCL; subst.

    eapply pesim_dance
    ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
    ; clear RLa0.
    - clear SV HV FIN; rename mrf into mrf0.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Slpc].
        eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq'.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq'.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sjz].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - unfold mrf; tmext_simpl; eauto.
      }
      { depmcplus0. }

      eapply dance_O; s.
      instantiate (1 := fun _ _ _ => _); s; eauto.

    - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
      esplits; first 2 last; last first.
      + right; splits; eauto using wfo_le_n_add.
      + i; frames_dep 1 RLa XTRa cfas.
        frames_nodep RLa cfas.
        destruct ea'; clear SV HV FIN.
        destruct_tgt_real RLa.

        setcrf mrf'.
        set (s' := (mrf', P0, s, s3, s4)).
        assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
        { clear -HV; unfold s'.
          s; splits; tmext_simpl; auto.
        }
        exists s'; esplits; [exact EXT'|].

        rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
        { unfold s'; rr; esplits.
          - rr; esplits; eauto.
            + s; rr; esplits; eauto; clean.
              instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp |}).
              rrdes CPa; s in HP.
              unfold mrf'; rr; splits; s; tmext_simpl; auto.
            + eapply tm_refconfrel_mon
              ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
              ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
          - rr; esplits; eauto.
          - s; noemp; f_equal.
            lhs rewrite -pmplus_assoc; eauto.
          - noemp; auto.
        }

        inst_pairs; eapply IHer; eauto using ext_trans.
        * i; hexploit RSV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * i; hexploit RDV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * unfold s', mrf'; s; tmext_simpl; eauto.
        * unfold s', mrf'; s; tmext_simpl; eauto.
        * rewrite -(app_nil_l envb).
          eapply tm_ksim_mon; eauto using ext_trans.
          unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
        * s; econs; symmetry.
          rewrite Eladdr !map_app !pmcells_app !map_length; s.
          rewrite !pmcells_cons map_app pmcells_app map_length; s.
          pmplus_eq'.
      + rr; esplits; s; noemp; eauto.

  }

  { 

    eapply pesim_stepb_obsolete.
    { i; rrdes XTR RLb; clarify.
      rrdes RL; s in Ec; noemp; clarify; esplits.
      - econs 1; inst_pairs; eapply IRok.
        eapply IRifnz; s; rewrite INb; s; eauto.
      - inst_pairs; s; noemp; rr; esplits; eauto.
    }

    do_getVal x Eladdr X.

    dup VCL; apply tm_vclosDb in VCL; des; rrdes VCL; subst.

    eapply pesim_dance
    ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
    ; clear RLa0.
    - clear SV HV FIN; rename mrf into mrf0.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Slpc].
        eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq'.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sjnz].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - tmext_simpl; eauto.
        - eauto.
      }
      { depmcplus0. }

      eapply dance_O; s.
      instantiate (1 := fun _ _ _ => _); s; eauto.

    - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
      esplits; first 2 last; last first.
      + right; splits; eauto using wfo_le_n_add.
      + i; frames_dep 1 RLa XTRa cfas.
        frames_nodep RLa cfas.
        destruct ea'; clear SV HV FIN.
        destruct_tgt_real RLa.

        setcrf mrf'.
        set (s' := (mrf', P0, s, s3, s4)).
        assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
        { clear -HV; unfold s'.
          s; splits; tmext_simpl; auto.
        }
        exists s'; esplits; [exact EXT'|].

        rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
        { unfold s'; rr; esplits.
          - rr; esplits; eauto.
            + s; rr; esplits; eauto; clean.
              instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp |}).
              rrdes CPa; s in HP.
              unfold mrf'; rr; splits; s; tmext_simpl; auto.
            + eapply tm_refconfrel_mon
              ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
              ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
          - rr; esplits; eauto.
          - s; noemp; f_equal.
            lhs rewrite -pmplus_assoc; eauto.
          - noemp; auto.
        }

        inst_pairs; eapply IHel; eauto using ext_trans.
        * i; hexploit RSV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * i; hexploit RDV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * unfold s', mrf'; s; tmext_simpl; eauto.
        * unfold s', mrf'; s; tmext_simpl; eauto.
        * rewrite -(app_nil_l envb).
          eapply tm_ksim_mon; eauto using ext_trans.
          unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
        * omega.
        * s; econs; symmetry.
          rewrite plus_assoc (plus_comm _ 3); s.
          rewrite Eladdr !map_app !pmcells_app !map_length; s.
          rewrite !pmcells_cons map_app pmcells_app map_length; s.
          pmplus_eq'.
      + rr; esplits; s; noemp; eauto.

  }

Grab Existential Variables.
exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0).
Qed.



Lemma tr_exp_correct_case: forall G x bx el er
    (IN: In (inl x) G)
    (IHel: tm_psim (inl (inr bx) :: G)
            (fun imports => map enc (tr_exp el imports (bx :: tvars G) (kvars G)))
            el)
    (IHer: tm_psim (inl (inr bx) :: G)
            (fun imports => map enc (tr_exp er imports (bx :: tvars G) (kvars G)))
            er),
  tm_psim G
    (fun imports => map enc (tr_exp (IEcase x bx el er) imports (tvars G) (kvars G)))
    (IEcase x bx el er).
Proof.
  i; hexploit getVal_length; i; des.
  econs; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  specialize (IHel i); rrdes IHel.
  specialize (IHer i); rrdes IHer.
  hexploit (values_for_var RSV RDV IN); i; des.

  tm_prepare.

  destruct (LEM (exists vb', vb = IVinl vb' \/ vb = IVinr vb'))
  ; last first.
  { 
    pfold; econs; i.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    eapply esim_err.
    - instantiate (1:=(None, (envb, IEcase x bx el er))).
      econs 1; eapply IRerr; vauto; i; invc FH
      ; rewrite INb in GET; clarify; eauto 10.
    - eauto.
  }
  des; subst.

  { 

    eapply pesim_stepb_obsolete.
    { i; rrdes XTR RLb; clarify.
      rrdes RL; s in Ec; noemp; clarify; esplits.
      - econs 1; inst_pairs; eapply IRok.
        eapply IRcasel; s; rewrite INb; s; eauto.
      - inst_pairs; s; noemp; rr; esplits; eauto.
    }

    do_getVal x Eladdr X.

    
    dup VCL; apply tm_vclosDb in VCL; des; rrdes VCL.
    dup CPa; rrdes CPa; clean; apply HP in VCL; s in VCL; clear HP ST; des.

    eapply pesim_dance
    ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
    ; clear RLa0.
    - clear SV HV FIN; rename mrf into mrf0.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - eauto.
      }
      { depmcplus0. }

      

      dance_init.
      { eapply Sok; [s; i; omega|eapply Salc].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - eauto.
      }
      { depmcplus1; eauto using alloc_valid. }
      s; setnewloc.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Ssto].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - rewrite !pmcells_cons -!pmplus_assoc (pmplus_swap {_|->_}).
          replace (S newloc) with (add_offset (mrf r_clo) o_one)
          ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - rrdes VCL1; econs; eapply pmread_plus; eauto.
          unfold mrf; s; tmext_simpl.
          rewrite RD; s; noemp; pmplus_eq'.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Ssto].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - rewrite !pmcells_cons -!pmplus_assoc.
          replace newloc with (add_offset (mrf r_clo) o_zero)
          ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }
      fold newloc.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - econs; eapply pmread_plus; eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Slpc].
        eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq'.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - rrdes VCL; econs; eapply pmread_plus; eauto.
          unfold mrf; s; tmext_simpl.
          rewrite RD; s; noemp; pmplus_eq'.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sjz].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - unfold mrf; tmext_simpl; eauto.
      }
      { depmcplus0. }

      eapply dance_O; s.
      instantiate (1 := fun _ cf _ _ _ _ _ _ _ _ => depmach1 cf); s; eauto.

    - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
      esplits; first 2 last; last first.
      + right; splits; eauto using wfo_le_n_add.
      + i; frames_dep 1 RLa XTRa cfas.
        frames_nodep RLa cfas.
        destruct ea'; clear SV HV FIN.
        destruct_tgt_real RLa.
        setnewloc.

        setcrf mrf'.
        set (s' := (mrf', P0, s, s3, (fst s4, {newloc |=> [va'; mrf r_env]} ## snd s4))).
        assert (VLD': pmvalid ({newloc |=> [va'; mrf r_env]} ## snd s4)).
        { decomp_pmvalid HV; eauto. }
        assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
        { clear -HV VLD'; unfold s'.
          s; splits; tmext_simpl; auto.
          i; eapply pmread_plus; eauto; pmplus_eq'.
        }
        exists s'; esplits; [exact EXT'|].

        rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
        { unfold s'; rr; esplits.
          - rr; esplits; eauto.
            + s; rr; esplits; eauto; clean.
              instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp |}).
              rrdes CPa0; s in HP.
              unfold mrf'; rr; splits; s; tmext_simpl; auto.
            + eapply tm_refconfrel_mon
              ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
              ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
          - rr; esplits; eauto.
          - s; noemp; f_equal.
            lhs rewrite pmplus_swap -pmplus_assoc; eauto.
          - noemp; auto.
        }

        inst_pairs; eapply IHel; eauto using ext_trans.
        * instantiate (1 := (inl (inr bx), vb') :: envb).
          instantiate (1 := senva).
          i; apply in_consD in IN0; des; vauto.
          rewrite list_assoc_cons_neq; [|by ii; vauto].
          hexploit RSV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * instantiate (1 := (bx, va') :: denva).
          i; apply in_consD in IN0; des.
          { clarify; esplits; eauto.
            eapply vclos_mon; eauto using (gk_mon GK).
          }
          rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN1; vauto]. 
          rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN1; vauto]. 
          hexploit RDV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * unfold s', mrf'; s; tmext_simpl; esplits
          ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
          eapply reprlist_mon; eauto.
          i; eapply pmread_plus; eauto; pmplus_eq'.
        * s; rewrite ORD; auto.
        * unfold s', mrf'; s; tmext_simpl; eauto.
        * s; rewrite cons_app.
          eapply tm_ksim_mon; eauto using ext_trans.
          unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
        * s; econs; symmetry.
          rewrite Eladdr !pmcells_cons !map_app !pmcells_app !map_length; s.
          rewrite !pmcells_cons map_app pmcells_app map_length; s.
          pmplus_eq'.
      + rr; esplits; s; noemp; eauto.

  }
  
  { 

    eapply pesim_stepb_obsolete.
    { i; rrdes XTR RLb; clarify.
      rrdes RL; s in Ec; noemp; clarify; esplits.
      - econs 1; inst_pairs; eapply IRok.
        eapply IRcaser; s; rewrite INb; s; eauto.
      - inst_pairs; s; noemp; rr; esplits; eauto.
    }

    do_getVal x Eladdr X.

    
    dup VCL; apply tm_vclosDb in VCL; des; rrdes VCL.
    dup CPa; rrdes CPa; clean; apply HP in VCL; s in VCL; clear HP ST; des.
    destruct n0; vauto.

    eapply pesim_dance
    ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
    ; clear RLa0.
    - clear SV HV FIN; rename mrf into mrf0.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - eauto.
      }
      { depmcplus0. }

      

      dance_init.
      { eapply Sok; [s; i; omega|eapply Salc].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - eauto.
      }
      { depmcplus1; eauto using alloc_valid. }
      s; setnewloc.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Ssto].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - rewrite !pmcells_cons -!pmplus_assoc (pmplus_swap {_|->_}).
          replace (S newloc) with (add_offset (mrf r_clo) o_one)
          ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_order; pmcells_fold.
        depmcplus0 ; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - rrdes VCL2; econs; eapply pmread_plus; eauto.
          unfold mrf; s; tmext_simpl.
          rewrite RD; s; noemp; pmplus_eq'.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Ssto].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - rewrite !pmcells_cons -!pmplus_assoc.
          replace newloc with (add_offset (mrf r_clo) o_zero)
          ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }
      fold newloc.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - econs; eapply pmread_plus; eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Slpc].
        eapply pmread_plus; eauto.
        rewrite Eladdr map_app pmcells_app map_length; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - rrdes VCL1; econs; eapply pmread_plus; eauto.
          unfold mrf; s; tmext_simpl.
          rewrite RD; s; noemp; pmplus_eq'.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sjnz].
        - eapply pmread_plus; eauto.
          rewrite Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; pmplus_eq''.
        - tmext_simpl; eauto.
        - eauto.
      }
      { depmcplus0. }

      eapply dance_O; s.
      instantiate (1 := fun _ cf _ _ _ _ _ _ _ _ => depmach1 cf); s; eauto.

    - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
      esplits; first 2 last; last first.
      + right; splits; eauto using wfo_le_n_add.
      + i; frames_dep 1 RLa XTRa cfas.
        frames_nodep RLa cfas.
        destruct ea'; clear SV HV FIN.
        destruct_tgt_real RLa.
        setnewloc.

        setcrf mrf'.
        set (s' := (mrf', P0, s, s3, (fst s4, {newloc |=> [va'; mrf r_env]} ## snd s4))).
        assert (VLD': pmvalid ({newloc |=> [va'; mrf r_env]} ## snd s4)).
        { decomp_pmvalid HV; eauto. }
        assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
        { clear -HV VLD'; unfold s'.
          s; splits; tmext_simpl; auto.
          i; eapply pmread_plus; eauto; pmplus_eq'.
        }
        exists s'; esplits; [exact EXT'|].

        rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
        { unfold s'; rr; esplits.
          - rr; esplits; eauto.
            + s; rr; esplits; eauto; clean.
              instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp |}).
              rrdes CPa0; s in HP.
              unfold mrf'; rr; splits; s; tmext_simpl; auto.
            + eapply tm_refconfrel_mon
              ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
              ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
          - rr; esplits; eauto.
          - s; noemp; f_equal.
            lhs rewrite pmplus_swap -pmplus_assoc; eauto.
          - noemp; auto.
        }

        inst_pairs; eapply IHer; eauto using ext_trans.
        * instantiate (1 := (inl (inr bx), vb') :: envb).
          instantiate (1 := senva).
          i; apply in_consD in IN0; des; vauto.
          rewrite list_assoc_cons_neq; [|by ii; vauto].
          hexploit RSV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * instantiate (1 := (bx, va') :: denva).
          i; apply in_consD in IN0; des.
          { clarify; esplits; eauto.
            eapply vclos_mon; eauto using (gk_mon GK).
          }
          rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN1; vauto]. 
          rewrite list_assoc_cons_neq; [|by ii; rewrite H in IN1; vauto]. 
          hexploit RDV; eauto; i; des; esplits; eauto.
          eapply vclos_mon; eauto using (gk_mon GK).
        * unfold s', mrf'; s; tmext_simpl; esplits
          ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
          eapply reprlist_mon; eauto.
          i; eapply pmread_plus; eauto; pmplus_eq'.
        * s; rewrite ORD; auto.
        * unfold s', mrf'; s; tmext_simpl; eauto.
        * s; rewrite cons_app.
          eapply tm_ksim_mon; eauto using ext_trans.
          unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
        * s; econs; symmetry.
          rewrite (plus_comm _ 4); s; rewrite -plus_comm.
          rewrite Eladdr !pmcells_cons !map_app !pmcells_app !map_length; s.
          rewrite !pmcells_cons map_app pmcells_app map_length; s.
          rewrite plus_comm; pmplus_eq'.
      + rr; esplits; s; noemp; eauto.

  }
Grab Existential Variables.
exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0).
Qed.



Lemma tr_exp_correct_let_contvar: forall G x ce k e
    (IHce: tm_psim (inl (inr x) :: G)
            (fun imports => map enc (tr_exp ce imports (x :: tvars G) (kvars G)))
            ce)
    (IHe: tm_psim (inr k :: G)
            (fun imports => map enc (tr_exp e imports (tvars G) (k :: kvars G)))
            e),
  tm_psim G
    (fun imports => map enc (tr_exp (IElet (IEcont x ce) (inr k) e) imports (tvars G) (kvars G)))
    (IElet (IEcont x ce) (inr k) e).
Proof.
  econs; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  specialize (IHce i); rrdes IHce.
  specialize (IHe (wfo_add i' i)); rrdes IHe. 

  tm_prepare.

  
  hexploit (tgt_stack CPa 2); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc in *; clear NST.

  eapply pesim_stepb_obsolete.
  { i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits.
    - econs 1; inst_pairs; eapply IRok.
      eapply IRletcont; s; rewrite INb; s; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
  }

  eapply pesim_dance; first (by eauto); clear RLa0.
  - clear SV HV FIN; rename mrf into mrf0.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Slpc]; read_code Eladdr. }
    { depmcplus0. }
    rename mrf into mrf0.       

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; [by read_code Eladdr| | |]; eauto. }
    { depmcplus0. }
    tmext_uniq.

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
      rewrite pmcells_cons -pmplus_assoc.
      replace (mrf0 r_sp)  with (add_offset (mrf r_sp) o_zero)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto]; [by read_code Eladdr|].
      rewrite 2!pmcells_cons -2!pmplus_assoc pmplus_swap.
      replace (S (mrf0 r_sp)) with (add_offset (mrf r_sp) o_one)
      ; last (by s; unfold mrf; tmext_simpl; eauto); eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; [by read_code Eladdr| | |]; eauto. }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := fun _ _ _ _ _ => _); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.

      

      setcrf mrf'.
      match goal with |- context[{_ |=> ?ws}] => set (cells := ws) end.
      set (s' := (mrf', P0, s, s3, (({mrf r_sp |=> cells} ## fst s4), snd s4))).
      assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { s; splits; tmext_simpl; auto. }
      esplits; eauto.

      rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s'; rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto; clean.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest0; chp := hp |}).
            rrdes CPa; s in HP; unfold mrf'; rr; splits; s; tmext_simpl; auto; s in ST.
            assert (XX := tgt_stack_transfer _ _ _ ST); s in XX; eauto.
          + eapply tm_refconfrel_mon
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
      - rr; esplits; eauto.
      - s; noemp; f_equal.
        + rewrite pmplus_swap; eauto.
        + rewrite -pmplus_assoc; auto.
      - noemp; auto.
      }

      inst_pairs; eapply IHe; safe eauto.
      * instantiate (1 := (inr k, IVclo envb (IEcont x ce)) :: envb).
        
        i; apply in_consD in IN; des; vauto.
        hexploit RSV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * i; apply in_consD in IN; des; vauto.
        hexploit RDV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * unfold s', mrf'; s; tmext_simpl; esplits; eauto.
      * s; eauto.
      * unfold s'; s.
        rewrite LST SP pmplus_swap plus_comm -LEN pmplus_assoc -pmcells_app; eauto.
      * rewrite app_length LEN; s; omega.
      * unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
      * { 
          s; econs.
          - s; eauto.
          - hexploit tm_ksim_mon; first (by eauto); last 3 first
            ; first (by eapply wfo_add_le2)
            ; first (by eauto); first 1 last
            ; first (by eapply ext_refl); eauto.
            { instantiate (1:=nil); s; auto. }
            i; simpl app; eapply tm_ksim_push; s; tmext_simpl; eauto.
            + unfold mrf'; s; tmext_simpl; omega.
            + decomp_pmvalid SV; auto.
          - unfold s', mrf'; s; tmext_simpl.
            rewrite plus_comm plus_minus_eq; [|omega]; s.
            decomp_pmvalid SV; eapply pmread_plus; eauto.
            unfold cells; rewrite pmcells_cons; pmplus_eq'.
          - unfold s', mrf'; s; tmext_simpl.
            rewrite (plus_comm _ 2) plus_minus_eq; [|s; omega]; rewrite plus_comm; s.
            decomp_pmvalid SV; eapply pmread_plus; eauto.
            unfold cells; rewrite 2!pmcells_cons; pmplus_eq'.
          - rewrite list_assoc_cons_eq; eauto.
            
          - i; invc RETa RETb.
            apply pesim_gknow; i.
            destruct s'0 as [[[[]]]]; destruct s7; clean.
            ren SV; ren HV; ren FIN; tm_prepare.

            eapply pesim_stepb_obsolete.
            { i; rrdes XTR RLb; clarify.
              rrdes RL; s in Ec; noemp; clarify; esplits.
              - econs 1; inst_pairs; eapply IRok.
                eapply IRret; eauto.
              - inst_pairs; s; noemp; rr; esplits; eauto.
            }

            hexploit (@pmplus_from_read _ _ _ (snd s4) h1).
            { i; rrdes EXT0; apply EXT1; s; auto. }
            { split_pmvalid HV0; eauto. }
            i; des; subst h1.

            eapply pesim_dance; first (by eauto); clear RLa0 RLa1. 
            { clear SV SV0 HV HV0 FIN FIN0; rename mrf into mrf1.

              dance_init.
              { eapply Sok; [s; i; omega|eapply Sld].
                - eapply pmread_plus; eauto.
                  rewrite Eladdr !map_app; s.
                  rewrite plus_comm; s; rewrite plus_comm !pmcells_cons.
                  rewrite pmcells_app map_length !pmcells_cons; s; pmplus_eq''.
                - eauto.
              }
              { depmcplus0. }
              rename mrf into mrf0.

              dance_init.
              { eapply Sok; [s; i; omega|eapply Sld].
                - eapply pmread_plus; eauto.
                  rewrite Eladdr !map_app; s.
                  rewrite plus_comm; s; rewrite plus_comm !pmcells_cons.
                  rewrite pmcells_app map_length !pmcells_cons; s; pmplus_eq''.
                - eauto.
              }
              { depmcplus0. }

              dance_init.
              { eapply Sok; [s; i; omega|eapply Salc].
                - eapply pmread_plus; eauto.
                  rewrite Eladdr !map_app; s.
                  rewrite plus_comm; s; rewrite plus_comm !pmcells_cons.
                  rewrite pmcells_app map_length !pmcells_cons; s; pmplus_eq''.
                - eauto.
              }
              { depmcplus1; eauto using alloc_valid. }
              s; setnewloc.

              dance_init.
              { eapply Sok; [s; i; omega|eapply Ssto].
                - eapply pmread_plus; eauto.
                  rewrite Eladdr !map_app; s.
                  rewrite plus_comm; s; rewrite plus_comm !pmcells_cons.
                  rewrite pmcells_app map_length !pmcells_cons; s; pmplus_eq''.
                - rewrite pmcells_cons -pmplus_assoc.
                  replace newloc with (add_offset (mrf r_env) o_zero)
                  ; last (by s; unfold mrf; tmext_simpl; auto); eauto.
              }
              { s; tmext_simpl; fold newloc; pmcells_fold.
                depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
              }

              dance_init.
              { eapply Sok; [s; i; omega|eapply Ssto].
                - eapply pmread_plus; eauto.
                  rewrite Eladdr !map_app; s.
                  rewrite plus_comm; s; rewrite plus_comm !pmcells_cons.
                  rewrite pmcells_app map_length !pmcells_cons; s; pmplus_eq''.
                - rewrite 2!pmcells_cons -2!pmplus_assoc (pmplus_swap {_|->_}).
                  replace (S newloc) with (add_offset (mrf r_env) o_one)
                  ; last (by s; unfold mrf; tmext_simpl; auto); eauto.
              }
              { s; tmext_simpl; fold newloc; pmcells_order; pmcells_fold.
                depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
              }

              eapply dance_O.
              instantiate (1 := fun _ _ cf _ _ => depmach1 cf); s; eauto.
            }

            i; rrdes XTRb RLb0; clarify.
            esplits; first 1 last.
            { right; split; eauto using wfo_le_n_add. }
            { eauto. }
    
            i; frames_dep 2 RLa XTRa cfas.
            frames_nodep RLa cfas.

            destruct ea'; clear SV HV FIN.
            destruct_tgt_real RLa.

            setnewloc.
            setcrf mrf''.
            set (s'' := (mrf'', P1, s5, s6,
                         (s7, {newloc |=> [mrf0 r_arg; mrf0 r_env]} ## snd s4 ## h''))).

            assert (VLD'': pmvalid ({newloc |=> [mrf0 r_arg; mrf0 r_env]} ## snd s4 ## h'')).
            { clear -HV.
              rewrite -!(pmplus_swap hp0) -!(pmplus_swap refha0) in HV.
              eapply pmvalid_plusD, pmvalid_plusD; eauto.
            }
            assert (EXT'': ext (wlift mylw) s'' (mrf0, P1, s5, s6, (s7, snd s4 ## h''))).
            { clear -VLD''.
              s; splits; tmext_simpl; auto.
              i; eapply pmread_plus; eauto.
              set (tmp := snd s4 ## _); pmplus_eq'.
            }

            exists s''; esplits; [exact EXT''|].
            rewrite tgt_csplit mid_csplit.
            eapply pesim_release; last first; try by rr; s; eauto.
            { unfold s''.
              rr; esplits.
              - rr; esplits; eauto.
                + s; rr; esplits; eauto. clean.
                  instantiate (1 := {| cpc := Nothing; crf := Just mrf''; cst := freest; chp := hp0 |}).
                  rrdes CPa0; s in HP.
                  unfold mrf''; rr; splits; s; tmext_simpl; auto.
                + eapply tm_refconfrel_mon
                  ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
                  ; s; i; eapply gk_mon; eauto; rrdes EXT''; rr; eauto.
              - rr; esplits; eauto.
              - s; noemp; f_equal.
                rewrite pmplus_swap -pmplus_assoc; eauto.
              - noemp; auto.
            }

            assert (S7: s7 = fst s4).
            { unfold s''; s; unfold s' in LST0; s in LST0.
              rewrite LST SP0 in LST0; unfold s', mrf' in LST0; s in LST0.
              tmext_simplH LST0; rewrite plus_comm plus_minus_eq in LST0; [|omega].
              s in LST0; rewrite (pmplus_comm s7) in LST0.
              eapply pmcells_uniq in LST0; eauto.
              * des; rewrite -LST1; eauto.
              * rewrite LST in SV0; fold cells in SV0.
                rewrite pmplus_swap in SV0; eapply pmvalid_plusD; eauto.
            }
            hide $ S7.          

            eapply IHce; eauto using ext_trans.
            + i; apply in_consD in IN; des; vauto.
              hexploit RSV; eauto; i; des; esplits; eauto.
              eapply vclos_mon; eauto using (gk_mon GK0), ext_trans.
            + instantiate (1 := ((x, mrf0 r_arg)) :: denva).
              i; apply in_consD in IN; des
              ; first (clarify; esplits; eauto
                       ; eapply vclos_mon; eauto using (gk_mon GK0)).
              rewrite list_assoc_cons_neq; [|by ii; vauto].
              rewrite list_assoc_cons_neq; [|by ii; vauto].
              hexploit RDV; eauto; i; des; esplits; eauto.
              eapply vclos_mon; eauto using (gk_mon GK0), ext_trans.
            + unfold s'', mrf''; s; tmext_simpl; esplits.
              * eapply pmread_plus; eauto.
                rewrite !pmcells_cons; pmplus_eq'.
              * eapply pmread_plus; eauto.
                rewrite !pmcells_cons; pmplus_eq'.
              * rewrite ENV0; eapply reprlist_mon; eauto.
                rewrite -ENV0; i; eapply pmread_plus; eauto; pmplus_eq'.
            + s; f_equal; auto.
            + unfold s''; s; rewrite S7; eauto.
            + unfold s'', mrf''; s; tmext_simpl; rewrite SP0.
              unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
            + s; rewrite cons_app.
              eapply tm_ksim_mon; eauto using ext_trans.
              unfold s'', mrf''; s; tmext_simpl; rewrite SP0.
              unfold s', mrf'; s; tmext_simpl; rewrite SP; omega.
            + s; econs; symmetry.
              rewrite plus_comm; s; rewrite -plus_comm.
              rewrite Eladdr !pmcells_cons map_app pmcells_app map_length.
              s; rewrite !pmcells_cons; pmplus_eq'.
        }
      * s; econs; symmetry.
        rewrite Eladdr !pmcells_cons map_app pmcells_app map_length.
        s; rewrite !pmcells_cons; pmplus_eq'.
    + rr; esplits; s; noemp; eauto.
Grab Existential Variables.
exact (wfo_nat 0).
Qed.



Lemma tr_exp_correct_let_termvar: forall G ae bx e
    (IHae: tm_bsim G
             (fun imports => map enc (tr_exp ae imports (tvars G) (kvars G)))
             ae)
    (IHe: tm_psim (inl (inr bx) :: G)
            (fun imports => map enc (tr_exp e imports (bx :: tvars G) (kvars G)))
            e),
  tm_psim G
    (fun imports => map enc (tr_exp (IElet ae (inl bx) e) imports (tvars G) (kvars G)))
    (IElet ae (inl bx) e).
Proof.
  econs; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  specialize (IHe i).
  rrdes IHae IHe.
  
  
  
  
  
  
  
  
  
  
  





Arguments cplus _ _ _ : simpl nomatch.
  tm_prepare.
  eapply pesim_dance; first (by eauto); clear RLa0.
  - clear SV HV FIN.
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app; s.
        rewrite ->?pmcells_cons; do 2 pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto].
      - eapply pmread_plus; eauto.
        rewrite Eladdr map_app; s.
        rewrite ->?pmcells_cons; pmplus_eq''.
      - rewrite 2!pmcells_cons -!pmplus_assoc (pmplus_swap {_|->_}).
        replace (S newloc) with (add_offset (mrf r_arg) o_one)
        ; last (by s; unfold mrf; tmext_simpl; auto); eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    eapply dance_O; instantiate (1 := fun _ cf _ => depmach1 cf); s; eauto.

  - i; rrdes XTRb RLb0; clarify.
    destruct ee; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.
      setnewloc.
      

      set (mrf' := [newloc//r_arg]>>[2//r_aux]>>mrf).
      set (s' := (mrf', P0, s, s3, (fst s4, {S newloc |-> mrf r_env} ## snd s4))).

      assert (VLD': pmvalid ({S newloc |-> mrf r_env} ## snd s4)).
      { clear -HV; rewrite -!pmplus_assoc in HV; ss; noemp.
        decomp_pmvalid HV; eauto.
      }

      


      assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
      { clear -HV VLD'; unfold s'.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto.
        rewrite ->pmplus_comm; eauto.
      }

      exists s'; split; first by auto.
      match goal with |- pesim _ _ _ _ _ _ _ _ ?cfg _ => replace cfg with
        (@cplus tgt_lsi
               {| cpc := Just (S (S (S mpc)));
                  crf := Nothing;
                  cst := pmemp;
                  chp := {mrf' r_arg |-> 0} |}
               {| cpc := Nothing;
                  crf := Just mrf';
                  cst := freest ## fst s4;
                  chp := {S newloc |-> mrf r_env} ## hp ## refha ## snd s4 |})
      end; idtac; last first.   
      { s; noemp; unfold mrf'; tmext_simpl; pmcells_fold; auto. }

      
      
      
      
      
      
      
      
      
      
      
      
      
      
      

      eapply IHae; eauto using ext_trans.
      * i; hexploit RSV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * i; hexploit RDV; eauto; i; des; esplits; eauto.
        eapply vclos_mon; eauto using (gk_mon GK).
      * unfold s', mrf'; s; tmext_simpl.
        eapply reprlist_mon; eauto.
        i; eapply pmread_plus; [by eauto|rhs rewrite pmplus_comm|]; eauto.
      * unfold s'; s; rewrite Eladdr; eexists; symmetry.
        rewrite !pmcells_cons map_app pmcells_app; pmplus_eq'.
      * {                       
        i; unfold s', mrf' in *; destruct s'0 as [[[[]]]]; clean.
        tmext_simplH ARG.
        dup EXT0; rrdes EXT0. rrdes EXT2. s in HP.
        clear SV HV FIN. rename ST into ST0; rename HP into HP0.

        tm_prepare.

        eapply pesim_dance; first (by eauto); clear RLa0.

        - clear SV HV FIN FIN0.
          dance_init.
          { s; setnewloc.
            rewrite plus_comm; s.
            eapply Sok; [s; i; omega|eapply Sld].
            - eapply pmread_plus; first (eapply HP0); eauto.
              + eapply pmread_plus; eauto.
                rewrite Eladdr map_app !pmcells_cons pmcells_app; s.
                rewrite ->?pmcells_cons; pmplus_eq''.
              + pmplus_eq'.
            - eauto.
          }
          { depmcplus0. }

          eapply dance_O; instantiate (1 := fun _ => _); s; eauto.

        - i; rrdes XTRb RLb0; clarify.
          esplits; first 2 last; last first.
          + right; splits; eauto using wfo_le_n_add.
          + i; frames_nodep RLa cfas.
            destruct ea'; clear SV HV FIN.
            destruct_tgt_real RLa.
            setnewloc.
            exists ([mrf0 r_arg//r_env]>>mrf0, P1, s5, s6, s7).
            split; [by s; auto|].
            esplits; eauto.
            rewrite tgt_csplit.
            eapply pesim_release; last first; try by rr; s; eauto.
            { rrdes CPa0 CPa.
              rrdes EXT1 EXT1 EXT1 EXT1.  s in RF0. s in RF1.
              rr; esplits; last by eauto.
              rr; esplits; eauto.
              - s; rr; esplits; eauto.
                instantiate (1 := mkconf _ _ _ _).
                rr; esplits; s; tmext_simpl; eauto.
                
              - eapply tm_refconfrel_mon
                ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
                ; s; i; eapply gk_mon; eauto; rr; s; eauto.
              - rr; esplits; eauto.
              - s; noemp; rewrite -pmplus_assoc; eauto.
            }

            assert (EXT'': ext (wlift mylw) ([mrf0 r_arg//r_env]>>mrf0, P1, s5, s6, s7)
                                            (mrf, P0, s, s3, s4)).
            { clear -EXT1 VLD'; rr in EXT1; ss; des; tmext_simplH RF; tmext_simplH RF0.
              splits; eauto.
              i; apply HP; eapply pmread_plus; eauto; pmplus_eq'.
            }

            hexploit (pmplus_from_read HP0); eauto; i; des.

            inst_pairs; eapply IHe; eauto.
            * eapply ext_trans; last first; eauto.
            * instantiate (1 := (inl (inr bx), vb) :: e0).
              i; apply in_consD in IN; des; vauto.
              hexploit RSV; eauto; i; des; esplits; eauto.
              eapply vclos_mon; eauto using (gk_mon GK).
            * instantiate (1 := (bx, va) :: denva).
              i; apply in_consD in IN; des.
              { clarify; esplits; eauto.
                eapply vclos_mon; eauto; first by rr; esplits; s; eauto.
                i; eapply gk_mon; eauto.
                rr; esplits; s; eauto.
              }
              rewrite list_assoc_cons_neq; [|by ii; vauto].
              rewrite list_assoc_cons_neq; [|by ii; vauto].
              hexploit RDV; eauto; i; des; esplits; eauto.
              eapply vclos_mon; eauto using (gk_mon GK).
            * s; tmext_simpl; esplits; eauto.
              { eapply HP0; rewrite ARG; eapply pmread_plus; eauto. }
              eapply reprlist_mon; eauto.
              i; eapply HP0; eapply pmread_plus; eauto.  rewrite ->pmplus_comm; eauto.
            * s; f_equal; eauto.
            * s; tmext_simpl; rewrite ST0; eauto.
            * rrdes EXT0 EXT0 EXT0; s in RF; tmext_simplH RF.
              s; tmext_simpl; rewrite -RF SP plus_0_r; s; auto.
            * s; rewrite -(app_nil_l (kvars _)) cons_app; tmext_simpl.
              rrdes EXT0 EXT0 EXT0; s in RF; tmext_simplH RF.
              eapply tm_ksim_mon; eauto.
              s; rewrite RF; tmext_simpl; auto.
            * s; econs; symmetry.
              rewrite X Eladdr !pmcells_cons map_app pmcells_app.
              s; rewrite pmcells_cons; pmplus_eq'.
              
          + s; noemp; s in RLb1; eauto.
        }
      * { unfold s', mrf'.
          rr; esplits; last by eauto.
          - rr; esplits; eauto.
            + s; rr; esplits; eauto.
              match type of RLa0 with tgt_real ?cfg _ =>
                instantiate (1 := {| cpc := Nothing; crf := crf cfg; cst := freest; chp := hp |})
              end; s.
              s; rrdes CPa; s; s in HP; s in ST.
              rr; splits; s; tmext_simpl; auto.
            + s; eapply tm_refconfrel_mon; s; unfold extend
              ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
              ; s; i; eapply gk_mon; eauto
              ; rrdes EXT'; rr; eauto using extpub_in_ext.
          - rr; esplits; eauto.
          - s; noemp; rewrite pmplus_swap -!pmplus_assoc; eauto.
        }
    + s; noemp; rr; esplits; eauto.
Qed.



Lemma tr_exp_correct_ret: forall G xk xv
    (INxk: In (inr xk) G)
    (INxv: In (inl xv) G),
  tm_psim G (fun imports => map enc (tr_exp (IEret xk xv) imports (tvars G) (kvars G)))
            (IEret xk xv).
Proof.
  i; hexploit getVal_length; i; des.
  eexists; i; des; s in Eladdr. 
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  hexploit (values_for_var RSV RDV INxv); i; des; clear RSV RDV.
  eapply kvars_in in INxk; hexploit tm_ksimD; eauto; i; des.
  

  tm_prepare.

  

  do_getVal xv Eladdr X.

  eapply pesim_getCont.
  - omega.
  - i; eapply pmread_plus; eauto.
    rewrite Eladdr map_app cons_app3 map_app !pmcells_app map_length; s.
    pmplus_eq'.
  - s; eapply wfo_le_n_add.
  - tmext_simpl; eapply pmread_plus; eauto; pmplus_eq'.
  - rewrite plus_comm; tmext_simpl; eapply pmread_plus; eauto; pmplus_eq'.
  - tmext_simpl; rewrite SP.
    instantiate (1 := 2 * length (kvars G) + r r_sp - 2 * (1 + index xk (kvars G))).
    hexploit (list_index_some _ _ INxk); i; des.
    rewrite X0; s; eapply list_index_length in X0; omega.
  - eapply pesim_dance
    ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
    ; clear RLa0.
    + clear SV HV FIN; rename mrf into mrf0.
      dance_init.
      { eapply Sok; [s; i; omega|eapply Sjmp].
        - eapply pmread_plus; eauto.
          rewrite plus_comm Eladdr map_app pmcells_app map_length; s.
          rewrite ->?pmcells_cons; s; pmplus_eq''.
        - eauto.
      }
      { depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells. }

      eapply dance_O; instantiate (1 := depmach0); s; eauto.

    + i; rrdes XTRb RLb0; clarify.
      destruct ee; s in Ec; noemp; clarify.

      hexploit (list_index_some _ _ INxk); i; des.
      rewrite ->X0 in *; simpl withd in *; rename n0 into idx.
      eapply list_index_length in X0.
      hexploit stack_split; eauto; i; des.
      hide $ SPL POPPED USED.
      rewrite SPL pmcells_app -pmplus_assoc in LST.

      esplits; first 2 last; last first.
      * right; split; eauto using wfo_le_n_add.
      * i; frames_nodep RLa cfas.
        destruct ea'; clear SV HV FIN.
        destruct_tgt_real RLa.
        rewrite plus_0_r.       

        setcrf mrf'.
        set (s' := (mrf', P0, s, s3, ({r r_sp |=> used} ## oldst, snd s4))).
        assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
        { unfold s'; rr; esplits; s; eauto. }

        assert (SP1: mrf' r_sp = r r_sp + length used).
        { unfold mrf'; tmext_simpl; rewrite LU LEN; omega. }
        assert (SP2: mrf' r_sp = mrf r_sp - length popped).
        { rewrite SP1 SP LU LEN LF; omega. }

        exists s'; split; [by rr; split; s; eauto|].
        rewrite tgt_csplit; eapply pesim_release; last first; try by rr; s; eauto.
        { unfold mrf'.
          rr; esplits; last by eauto.
          - rr; esplits; [ | |by eauto|by eauto]; last first.
            + eapply tm_refconfrel_mon
              ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
              ; s; i; eapply gk_mon; eauto; rr; s; eauto.
            + rr; splits.
              * eapply tgt_confpred_pop with (popped:=popped); eauto.
                rewrite LST in SV; decomp_pmvalid SV; rewrite SP1 pmplus_comm; eauto.
              * eauto.
          - rr; esplits; eauto.
          - simpl cplus; noemp; f_equal.
            + tmext_simpl; rewrite -!pmplus_assoc (pmplus_swap {_ |=> _}).
              rewrite SP1 LST; rhs rewrite (pmplus_swap {_|=>_}); auto.
            + rewrite -pmplus_assoc; auto.
        }
        eapply ESIM; eauto.
        { unfold s', mrf'; s; tmext_simpl; rewrite plus_0_r SP; s; auto. }
        { s; rewrite LST -pmplus_assoc (pmplus_comm oldst) SP1; auto. }
        { unfold s', mrf'; s. tmext_simpl; auto. }
        { eapply vclos_mon; eauto.
          i; eapply gk_mon; eauto.
        }
        { eapply tgt_cquery_ret.
          unfold s', mrf'; s; tmext_simpl; eauto.
        }
        { s; inst_pairs; eapply mid_cquery_ret; eauto. }
      * s; noemp; rr; esplits; eauto.
Qed.
