(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List NPeano.
Require Import vgtac basis types lang_common lang_src_lsi lang_tgt.
Require Import model_untyped model gw_common gw_untyped gw.
Require Import cps opt_freshen opt_dce opt_commute opt_hoist opt_dedup opt_inline opt_contify codegen.
Require Import cps_correct codegen_correct vcomp_tms vcomp_mms compiler_common.

Set Implicit Arguments.

Global Set Bullet Behavior "Strict Subproofs".

Axiom le_exp: lang_mid.exp -> lang_mid.exp -> bool.




Definition compile (tfs: list modtf) (FI: list label) : lang_src.module -> lang_tgt.module :=
  codegen.tr_mod FI <*> apply_modtfs tfs FI <*> mcps.
  

Theorem compile_correct: forall tfs FI FO FI' FO' m
    (WF: mhastype FI' m FO')
    (EFI: FI' = map_pair id model.flext_of FI)
    (EFO: FO' = map_pair id model.flext_of FO)
    (TF: forall f (IN: In f tfs), <<WF: modtf_wf f>> /\ <<SIM: modtf_sim f>>),
  msim ts_gw FI FO (compile tfs (map fst FI) m) m /\
  tgt_wfmod (compile tfs (map fst FI) m) (map fst FO).
Proof.
  i; unfold compile; s; subst.
  hexploit mcps_correct; eauto; []; i; des.
  hexploit mcps_uniqvar; eauto; []; hexploit mcps_wfmod; eauto; []; i; des.
  hexploit apply_modtfs_correct; eauto; []; i; des.
  split.
  - eapply vcomp_tms; [eapply tr_mod_correct|eapply vcomp_mms_rtc]; eauto.
  - apply tr_mod_wf; eauto.
Qed.


Structure selection :=
  { do_inline: bool
  ; do_contify: bool
  ; do_dce: bool
  ; do_hoist: bool
  ; do_commute: bool
  ; do_dedup: bool
  }.

Definition do_all: selection :=
  Build_selection true true true true true true.

Definition selected_transformations sel : list modtf :=
  (if do_inline sel then [fun _ => freshen_mod <*> inline_mod] else nil) ++
  (if do_contify sel then [fun FI => contify_mod (map (inl<*>inl) FI) <*> freshen_mod] else nil) ++
  (if do_dce sel then [dce_mod] else nil) ++
  (if do_hoist sel then [hoist_mod] else nil) ++
  (if do_commute sel then [commute_mod le_exp] else nil) ++
  (if do_dedup sel then [dedup_mod] else nil) ++
  nil.

Lemma selected_transformations_correct: forall sel f
    (IN: In f (selected_transformations sel)),
  <<WF: modtf_wf f>> /\ <<SIM: modtf_sim f>>.
Proof.
  i.
  apply in_app_or in IN; des.
  { des_ifs; ss; des; clarify; auto using inline_freshen_correct. }
  apply in_app_or in IN; des.
  { des_ifs; ss; des; clarify; auto using freshen_contify_correct. }
  apply in_app_or in IN; des.
  { des_ifs; ss; des; clarify; auto using dce_correct. }
  apply in_app_or in IN; des.
  { des_ifs; ss; des; clarify; auto using hoist_correct. }
  apply in_app_or in IN; des.
  { des_ifs; ss; des; clarify; auto using commute_correct. }
  apply in_app_or in IN; des.
  { des_ifs; ss; des; clarify; auto using dedup_correct. }
  vauto.
Qed.


Definition pilsner sel :=
  compile (selected_transformations sel).

Theorem pilsner_correct: forall n FI FO FI' FO' m
    (WF: mhastype FI' m FO')
    (EFI: FI' = map_pair id model.flext_of FI)
    (EFO: FO' = map_pair id model.flext_of FO),
  <<SIM: msim ts_gw FI FO (pilsner n (map fst FI) m) m>> /\
  <<WF: tgt_wfmod (pilsner n (map fst FI) m) (map fst FO)>>.
Proof.
  i; eapply compile_correct; eauto using selected_transformations_correct.
Qed.


